from Dictionary.TxtWord import TxtWord
from Language.TurkishLanguage import TurkishLanguage

from MorphologicalAnalysis.FsmParse import FsmParse
from MorphologicalAnalysis.State import State


class Transition:
    __toState: State
    __with: str
    __withName: str
    __formationToCheck: str
    __toPos: str

    """
    Constructor of Transition class which takes a State, and three str as input. Then it
    initializes toState, with, withName and toPos variables with given inputs.

    PARAMETERS
    ----------
    with : str    
        String input.
    toState : State 
        State input.
    withName : str
        String input.
    toPos : str   
        String input.
    """

    def __init__(self, _with: str, toState=None, withName=None, toPos=None):
        self.__with = _with
        self.__toState = toState
        self.__withName = withName
        self.__toPos = toPos

    """
    Getter for the toState variable.

    RETURNS
    -------
    State
        toState variable.
    """

    def toState(self) -> State:
        return self.__toState

    """
    Getter for the toPos variable.

    RETURNS
    -------
    str
        toPos variable.
    """

    def toPos(self) -> str:
        return self.__toPos

    """
    The transitionPossibleForString method takes two str as inputs; currentSurfaceForm and realSurfaceForm. If the
    length of the given currentSurfaceForm is greater than the given realSurfaceForm, it directly returns true. If not,
    it takes a substring from given realSurfaceForm with the size of currentSurfaceForm. Then checks for the characters 
    of with variable.

    If the character of with that makes transition is C, it returns true if the substring contains c or ç.
    If the character of with that makes transition is D, it returns true if the substring contains d or t.
    If the character of with that makes transition is A, it returns true if the substring contains a or e.
    If the character of with that makes transition is K, it returns true if the substring contains k, g or ğ.
    If the character of with that makes transition is other than the ones above, it returns true if the substring
    contains the same character as with.

    PARAMETERS
    ----------
    currentSurfaceForm : str
        String input.
    realSurfaceForm : str   
        String input.
        
    RETURNS
    -------
    bool
        True when the transition is possible according to Turkish grammar, False otherwise.
    """

    def transitionPossibleForString(self, currentSurfaceForm: str, realSurfaceForm: str) -> bool:
        if len(currentSurfaceForm) == 0 or len(currentSurfaceForm) >= len(realSurfaceForm):
            return True
        searchString = realSurfaceForm[len(currentSurfaceForm):]
        for ch in self.__with:
            if ch == 'C':
                return 'c' in searchString or 'ç' in searchString
            elif ch == 'D':
                return 'd' in searchString or 't' in searchString
            elif ch == 'c' or ch == 'e' or ch == 'r' or ch == 'p' or ch == 'l' or ch == 'b' or ch == 'd' or ch == 'g' \
                    or ch == 'o' or ch == 'm' or ch == 'v' or ch == 'i' or ch == 'ü' or ch == 'z':
                return ch in searchString
            elif ch == 'A':
                return 'a' in searchString or 'e' in searchString
            elif ch == 'k':
                return 'k' in searchString or 'g' in searchString or 'ğ' in searchString
        return True

    """
    The transitionPossibleForParse method takes a FsmParse currentFsmParse as an input. It then checks some special cases;

    PARAMETERS
    ----------
    currentFsmParse : FsmParse
        Parse to be checked
        
    RETURNS
    -------
    bool
        True if transition is possible False otherwise
    """

    def transitionPossibleForParse(self, currentFsmParse: FsmParse) -> bool:
        if self.__with == "Ar" and currentFsmParse.getSurfaceForm().endswith("l") and \
                currentFsmParse.getWord().getName() != currentFsmParse.getSurfaceForm():
            return False
        if currentFsmParse.getVerbAgreement() is not None and currentFsmParse.getPossesiveAgreement() is not None and \
                self.__withName is not None:
            if currentFsmParse.getVerbAgreement() == "A3PL" and self.__withName == "^DB+VERB+ZERO+PRES+A1SG":
                return False
            if currentFsmParse.getVerbAgreement() == "A3SG" and (currentFsmParse.getPossesiveAgreement() == "P1SG" or
                                                                 currentFsmParse.getPossesiveAgreement() == "P2SG") \
                    and self.__withName == "^DB+VERB+ZERO+PRES+A1PL":
                return False
        return True

    def transitionPossibleForWord(self, root: TxtWord, fromState: State) -> bool:
        if root.isAdjective() and ((root.isNominal() and not root.isExceptional()) or root.isPronoun()) \
                and self.__toState.getName() == "NominalRoot(ADJ)" and self.__with == "0":
            return False
        if root.isAdjective() and root.isNominal() and self.__with == "^DB+VERB+ZERO+PRES+A3PL" \
                and fromState.getName() == "AdjectiveRoot":
            return False
        if root.isAdjective() and root.isNominal() and self.__with == "SH" and fromState.getName() == "AdjectiveRoot":
            return False
        if self.__with == "ki":
            return root.takesRelativeSuffixKi()
        if self.__with == "kü":
            return root.takesRelativeSuffixKu()
        if self.__with == "dHr":
            if self.__toState.getName() == "Adverb":
                return True
            else:
                return root.takesSuffixDIRAsFactitive()
        if self.__with == "Hr" and (
                self.__toState.getName() == "AdjectiveRoot(VERB)" or self.__toState.getName() == "OtherTense" or self.__toState.getName() == "OtherTense2"):
            return root.takesSuffixIRAsAorist()
        return True

    """
    The beforeLastVowel method takes a str stem as an input. It loops through the given stem and returns
    the second last vowel.

    PARAMETERS
    ----------
    stem : str
        String input.
        
    RETURNS
    -------
    str
        Vowel before the last vowel.
    """

    def beforeLastVowel(self, stem: str) -> str:
        last = "0"
        before = 1
        for i in range(len(stem) - 1, -1, -1):
            if TurkishLanguage.isVowel(stem[i]):
                if before == 1:
                    last = stem[i]
                    before = before - 1
                    continue
                return stem[i]
        return last

    """
    The lastVowel method takes a str stem as an input. It loops through the given stem and returns
    the last vowel.

    PARAMETERS
    ----------
    stem : str
        String input.
        
    RETURNS
    -------
    str
        The last vowel.
    """

    def lastVowel(self, stem: str) -> str:
        for i in range(len(stem) - 1, - 1, -1):
            if TurkishLanguage.isVowel(stem[i]):
                return stem[i]
        for i in range(len(stem) - 1, -1, -1):
            if "0" <= stem[i] <= "9":
                return stem[i]
        return "0"

    """
    The lastPhoneme method takes a str stem as an input. It then returns the last phoneme of the given stem.

    PARAMETERS
    ----------
    stem : str
        String input.
        
    RETURNS
    -------
    str
        The last phoneme.
    """

    def lastPhoneme(self, stem: str) -> str:
        if len(stem) == 0:
            return " "
        if stem[len(stem) - 1] != "'":
            return stem[len(stem) - 1]
        else:
            return stem[len(stem) - 2]

    """
    The withFirstChar method returns the first character of the with variable.

    RETURNS
    -------
    str
        The first character of the with variable.
    """

    def withFirstChar(self) -> str:
        if len(self.__with) == 0:
            return "$"
        if self.__with[0] != "'":
            return self.__with[0]
        elif len(self.__with) == 1:
            return self.__with[0]
        else:
            return self.__with[1]

    """
    The startWithVowelorConsonantDrops method checks for some cases. If the first character of with variable is "nsy",
    and with variable does not equal to one of the Strings; "ylA, ysA, ymHs, yDH, yken", it returns true. If

    Or, if the first character of with variable is 'A, H: or any other vowels, it returns true.

    RETURNS
    -------
    bool
        True if it starts with vowel or consonant drops, false otherwise.
    """

    def startWithVowelorConsonantDrops(self) -> bool:
        if TurkishLanguage.isConsonantDrop(self.withFirstChar()) and self.__with != "ylA" and self.__with != "ysA" \
                and self.__with != "ymHs" and self.__with != "yDH" and self.__with != "yken":
            return True
        if self.withFirstChar() == "A" or self.withFirstChar() == "H" or TurkishLanguage.isVowel(self.withFirstChar()):
            return True
        return False

    """
    The softenDuringSuffixation method takes a TxtWord root as an input. It checks two cases; first case returns
    true if the given root is nominal or adjective and has one of the flags "IS_SD, IS_B_SD, IS_SDD" and with variable
    equals o one of the Strings "Hm, nDAn, ncA, nDA, yA, yHm, yHz, yH, nH, nA, nHn, H, sH, Hn, HnHz, HmHz".

    And the second case returns true if the given root is verb and has the "F_SD" flag, also with variable starts with
    "Hyor" or equals one of the Strings "yHs, yAn, yA, yAcAk, yAsH, yHncA, yHp, yAlH, yArAk, yAdur, yHver, yAgel, yAgor,
    yAbil, yAyaz, yAkal, yAkoy, yAmA, yHcH, HCH, Hr, Hs, Hn, yHn", yHnHz, Ar, Hl").

    PARAMETERS
    ----------
    root : TxtWord
        TxtWord input.
        
    RETURNS
    -------
    bool
        True if there is softening during suffixation of the given root, False otherwise.
    """

    def softenDuringSuffixation(self, root: TxtWord) -> bool:
        if (root.isNominal() or root.isAdjective()) and root.nounSoftenDuringSuffixation() and \
                (self.__with == "Hm" or self.__with == "nDAn" or self.__with == "ncA" or self.__with == "nDA"
                 or self.__with == "yA" or self.__with == "yHm" or self.__with == "yHz" or self.__with == "yH"
                 or self.__with == "nH" or self.__with == "nA" or self.__with == "nHn" or self.__with == "H"
                 or self.__with == "sH" or self.__with == "Hn" or self.__with == "HnHz" or self.__with == "HmHz"):
            return True
        if root.isVerb() and root.verbSoftenDuringSuffixation() and \
                (self.__with.startswith("Hyor") or self.__with == "yHs" or self.__with == "yAn" or self.__with == "yA"
                 or self.__with == "yAcAk" or self.__with == "yAsH" or self.__with == "yHncA" or self.__with == "yHp"
                 or self.__with == "yAlH" or self.__with == "yArAk" or self.__with == "yAdur" or self.__with == "yHver"
                 or self.__with == "yAgel" or self.__with == "yAgor" or self.__with == "yAbil" or self.__with == "yAyaz"
                 or self.__with == "yAkal" or self.__with == "yAkoy" or self.__with == "yAmA" or self.__with == "yHcH"
                 or self.__with == "yHs" or self.__with == "HCH" or self.__with == "Hr" or self.__with == "Hs"
                 or self.__with == "Hn" or self.__with == "yHn" or self.__with == "yHnHz" or self.__with == "Ar"
                 or self.__with == "Hl"):
            return True
        return False

    """
    The makeTransition method takes a TxtWord root and s str stem as inputs. If given root is a verb,
    it makes transition with given root and stem with the verbal root state. If given root is not verb, it makes 
    transition with given root and stem and the nominal root state.

    PARAMETERS
    ----------
    root : TxtWord
        TxtWord input.
    stem : str
        String input.
        
    RETURNS
    -------
    str
        String type output that has the transition.
    """

    def makeTransitionNoStartState(self, root: TxtWord, stem: str) -> str:
        if root.isVerb():
            return self.makeTransition(root, stem, State("VerbalRoot", True, False))
        else:
            return self.makeTransition(root, stem, State("NominalRoot", True, False))

    def makeTransition(self, root: TxtWord, stem: str, startState: State) -> str:
        rootWord = root.getName() == stem or (root.getName() + "'") == stem
        formation = stem
        i = 0
        if self.__with == "0":
            return stem
        if (stem == "bu" or stem == "şu" or stem == "o") and rootWord and self.__with == "ylA":
            return stem + "nunla"
        if self.__with == "yA":
            if stem == "ben":
                return "bana"
        self.__formationToCheck = stem
        if rootWord and self.withFirstChar() == "y" and root.vowelEChangesToIDuringYSuffixation() \
                and self.__with[1] != "H":
            formation = stem[:len(stem) - 1] + "i"
            self.__formationToCheck = formation
        else:
            if rootWord and (self.__with == "Hl" or self.__with == "Hn") and root.lastIdropsDuringPassiveSuffixation():
                formation = stem[:len(stem) - 2] + stem[len(stem) - 1]
                self.__formationToCheck = stem
            else:
                if rootWord and root.showsSuRegularities() and self.startWithVowelorConsonantDrops() and \
                        not self.__with.startswith("y"):
                    formation = stem + 'y'
                    self.__formationToCheck = formation
                else:
                    if rootWord and root.duplicatesDuringSuffixation() and TurkishLanguage.isConsonantDrop(
                            self.__with[0]):
                        if self.softenDuringSuffixation(root):
                            if self.lastPhoneme(stem) == "p":
                                formation = stem[:len(stem) - 1] + "bb"
                            elif self.lastPhoneme(stem) == "t":
                                formation = stem[:len(stem) - 1] + "dd"
                        else:
                            formation = stem + stem[len(stem) - 1]
                        self.__formationToCheck = formation
                    else:
                        if rootWord and root.lastIdropsDuringSuffixation() and \
                                not startState.getName().startswith(
                                    "VerbalRoot") and not startState.getName().startswith("ProperRoot") \
                                and self.startWithVowelorConsonantDrops():
                            if self.softenDuringSuffixation(root):
                                if self.lastPhoneme(stem) == "p":
                                    formation = stem[:len(stem) - 2] + 'b'
                                elif self.lastPhoneme(stem) == "t":
                                    formation = stem[:len(stem) - 2] + 'd'
                                elif self.lastPhoneme(stem) == "ç":
                                    formation = stem[:len(stem) - 2] + 'c'
                            else:
                                formation = stem[:len(stem) - 2] + stem[len(stem) - 1]
                            self.__formationToCheck = stem
                        else:
                            if self.lastPhoneme(stem) == "p":
                                if self.startWithVowelorConsonantDrops() and rootWord and self.softenDuringSuffixation(
                                        root):
                                    formation = stem[:len(stem) - 1] + 'b'
                            elif self.lastPhoneme(stem) == "t":
                                if self.startWithVowelorConsonantDrops() and rootWord and self.softenDuringSuffixation(
                                        root):
                                    formation = stem[:len(stem) - 1] + 'd'
                            elif self.lastPhoneme(stem) == "ç":
                                if self.startWithVowelorConsonantDrops() and rootWord and self.softenDuringSuffixation(
                                        root):
                                    formation = stem[:len(stem) - 1] + 'c'
                            elif self.lastPhoneme(stem) == "g":
                                if self.startWithVowelorConsonantDrops() and rootWord and self.softenDuringSuffixation(
                                        root):
                                    formation = stem[:len(stem) - 1] + 'ğ'
                            elif self.lastPhoneme(stem) == "k":
                                if self.startWithVowelorConsonantDrops() and rootWord and root.endingKChangesIntoG() \
                                        and not root.isProperNoun():
                                    formation = stem[:len(stem) - 1] + 'g'
                                else:
                                    if self.startWithVowelorConsonantDrops() and (not rootWord or (
                                            self.softenDuringSuffixation(root) and (
                                            not root.isProperNoun() or startState.__str__() != "ProperRoot"))):
                                        formation = stem[:len(stem) - 1] + 'ğ'
                            self.__formationToCheck = formation
        if TurkishLanguage.isConsonantDrop(self.withFirstChar()) and not TurkishLanguage.isVowel(stem[len(stem) - 1]) \
                and (root.isNumeral() or root.isReal() or root.isFraction() or root.isTime() or root.isDate()
                     or root.isPercent() or root.isRange()) \
                and (root.getName().endswith("1") or root.getName().endswith("3") or root.getName().endswith("4")
                     or root.getName().endswith("5") or root.getName().endswith("8") or root.getName().endswith("9")
                     or root.getName().endswith("10") or root.getName().endswith("30") or root.getName().endswith("40")
                     or root.getName().endswith("60") or root.getName().endswith("70") or root.getName().endswith("80")
                     or root.getName().endswith("90") or root.getName().endswith("00")):
            if self.__with[0] == "'":
                formation = formation + "'"
                i = 2
            else:
                i = 1
        else:
            if (TurkishLanguage.isConsonantDrop(self.withFirstChar()) and TurkishLanguage.isConsonant(
                    self.lastPhoneme(stem))) or (rootWord and root.consonantSMayInsertedDuringPossesiveSuffixation()):
                if self.__with[0] == "'":
                    formation = formation + "'"
                    if root.isAbbreviation():
                        i = 1
                    else:
                        i = 2
                else:
                    i = 1
        while i < len(self.__with):
            if self.__with[i] == "D":
                formation = self.resolveD(root, formation)
            elif self.__with[i] == "A":
                formation = self.resolveA(root, formation, rootWord)
            elif self.__with[i] == "H":
                if self.__with[0] != "'":
                    formation = self.resolveH(root, formation, i == 0, self.__with.startswith("Hyor"), rootWord)
                else:
                    formation = self.resolveH(root, formation, i == 1, False, rootWord)
            elif self.__with[i] == "C":
                formation = self.resolveC(formation)
            elif self.__with[i] == "S":
                formation = self.resolveS(formation)
            elif self.__with[i] == "Ş":
                formation = self.resolveSh(formation)
            else:
                if i == len(self.__with) - 1 and self.__with[i] == "s":
                    formation += "ş"
                else:
                    formation += self.__with[i]
            self.__formationToCheck = formation
            i = i + 1
        return formation

    def resolveD(self, root: TxtWord, formation: str) -> str:
        if root.isAbbreviation():
            return formation + 'd'
        if "0" <= self.lastPhoneme(self.__formationToCheck) <= "9":
            if self.lastPhoneme(self.__formationToCheck) == "3" or self.lastPhoneme(self.__formationToCheck) == "4" \
                    or self.lastPhoneme(self.__formationToCheck) == "5":
                return formation + 't'
            elif self.lastPhoneme(self.__formationToCheck) == "0":
                if root.getName().endswith("40") or root.getName().endswith("60") or root.getName().endswith("70"):
                    return formation + 't'
                else:
                    return formation + 'd'
            else:
                return formation + 'd'
        else:
            if TurkishLanguage.isSertSessiz(self.lastPhoneme(self.__formationToCheck)):
                return formation + 't'
            else:
                return formation + 'd'

    def resolveA(self, root: TxtWord, formation: str, rootWord: bool):
        if root.isAbbreviation():
            return formation + 'e'
        if "0" <= self.lastVowel(self.__formationToCheck) <= "9":
            if self.lastVowel(self.__formationToCheck) == "6" or self.lastVowel(self.__formationToCheck) == 9:
                return formation + 'a'
            elif self.lastVowel(self.__formationToCheck) == "0":
                if root.getName().endswith("10") or root.getName().endswith("30") or root.getName().endswith("40") \
                        or root.getName().endswith("60") or root.getName().endswith("90"):
                    return formation + 'a'
                else:
                    return formation + 'e'
            else:
                return formation + 'e'
        if TurkishLanguage.isBackVowel(self.lastVowel(self.__formationToCheck)):
            if root.notObeysVowelHarmonyDuringAgglutination() and rootWord:
                return formation + 'e'
            else:
                return formation + 'a'
        if TurkishLanguage.isFrontVowel(self.lastVowel(self.__formationToCheck)):
            if root.notObeysVowelHarmonyDuringAgglutination() and rootWord:
                return formation + 'a'
            else:
                return formation + 'e'
        if root.isNumeral() or root.isFraction() or root.isReal():
            if root.getName().endswith("6") or root.getName().endswith("9") or root.getName().endswith("10") or \
                    root.getName().endswith("30") or root.getName().endswith("40") or root.getName().endswith("60") \
                    or root.getName().endswith("90"):
                return formation + 'a'
            else:
                return formation + 'e'
        return formation

    def resolveH(self, root: TxtWord, formation: str, beginningOfSuffix: bool, specialCaseTenseSuffix: bool, rootWord: bool):
        if root.isAbbreviation():
            return formation + 'i'
        if beginningOfSuffix and TurkishLanguage.isVowel(self.lastPhoneme(self.__formationToCheck)) and not specialCaseTenseSuffix:
            return formation
        if specialCaseTenseSuffix:
            if rootWord:
                if root.vowelAChangesToIDuringYSuffixation():
                    if TurkishLanguage.isFrontRoundedVowel(self.beforeLastVowel(self.__formationToCheck)):
                        return formation[:len(formation) - 1] + 'ü'
                    if TurkishLanguage.isFrontUnroundedVowel(self.beforeLastVowel(self.__formationToCheck)):
                        return formation[:len(formation) - 1] + 'i'
                    if TurkishLanguage.isBackRoundedVowel(self.beforeLastVowel(self.__formationToCheck)):
                        return formation[:len(formation) - 1] + 'u'
                    if TurkishLanguage.isBackUnroundedVowel(self.beforeLastVowel(self.__formationToCheck)):
                        return formation[:len(formation) - 1] + 'ı'
            if TurkishLanguage.isVowel(self.lastPhoneme(self.__formationToCheck)):
                if TurkishLanguage.isFrontRoundedVowel(self.beforeLastVowel(self.__formationToCheck)):
                    return formation[:len(formation) - 1] + 'ü'
                if TurkishLanguage.isFrontUnroundedVowel(self.beforeLastVowel(self.__formationToCheck)):
                    return formation[:len(formation) - 1] + 'i'
                if TurkishLanguage.isBackRoundedVowel(self.beforeLastVowel(self.__formationToCheck)):
                    return formation[:len(formation) - 1] + 'u'
                if TurkishLanguage.isBackUnroundedVowel(self.beforeLastVowel(self.__formationToCheck)):
                    return formation[:len(formation) - 1] + 'ı'
        if TurkishLanguage.isFrontRoundedVowel(self.lastVowel(self.__formationToCheck)) or \
                (TurkishLanguage.isBackRoundedVowel(self.lastVowel(self.__formationToCheck))
                 and root.notObeysVowelHarmonyDuringAgglutination()):
            return formation + 'ü'
        if TurkishLanguage.isFrontUnroundedVowel(self.lastVowel(self.__formationToCheck)) or \
                (self.lastVowel(self.__formationToCheck) == 'a' and root.notObeysVowelHarmonyDuringAgglutination()):
            return formation + 'i'
        if TurkishLanguage.isBackRoundedVowel(self.lastVowel(self.__formationToCheck)):
            return formation + 'u'
        if TurkishLanguage.isBackUnroundedVowel(self.lastVowel(self.__formationToCheck)):
            return formation + 'ı'
        if root.isNumeral() or root.isFraction() or root.isReal():
            if root.getName().endswith("6") or root.getName().endswith("40") or root.getName().endswith("60") \
                    or root.getName().endswith("90"):
                return formation + 'ı'
            else:
                if root.getName().endswith("3") or root.getName().endswith("4") or root.getName().endswith("00"):
                    return formation + 'ü'
                else:
                    if root.getName().endswith("9") or root.getName().endswith("10") or root.getName().endswith("30"):
                        return formation + 'u'
                    else:
                        return formation + 'i'
        return formation

    """
    The resolveC method takes a str formation as an input. If the last phoneme is on of the "çfhkpsşt", it
    concatenates given formation with 'ç', if not it concatenates given formation with 'c'.

    PARAMETERS
    ----------
    formation : str
        String input.
        
    RETURNS
    -------
    str
        Resolved String.
    """
    def resolveC(self, formation: str) -> str:
        if TurkishLanguage.isSertSessiz(self.lastPhoneme(self.__formationToCheck)):
            return formation + 'ç'
        else:
            return formation + 'c'

    """
    The resolveS method takes a str formation as an input. It then concatenates given formation with 's'.

    PARAMETERS
    ----------
    formation : str
        String input.
        
    RETURNS
    -------
    str
        Resolved String.
    """
    def resolveS(self, formation: str) -> str:
        return formation + 's'

    """
    The resolveSh method takes a str formation as an input. If the last character is a vowel, it concatenates
    given formation with ş, if the last character is not a vowel, and not 't' it directly returns given formation, 
    but if it is equal to 't', it transforms it to 'd'.

    PARAMETERS
    ----------
    formation : str
        String input.
        
    RETURNS
    -------
    str
        Resolved String.
    """
    def resolveSh(self, formation: str) -> str:
        if TurkishLanguage.isVowel(formation[len(formation) - 1]):
            return formation + 'ş'
        else:
            if formation[len(formation) - 1] != 't':
                return formation
            else:
                return formation[len(formation) - 1] + 'd'

    """
    An overridden toString method which returns the with variable.

    RETURNS
    -------
    str
        With variable.
    """
    def __str__(self):
        return self.__with

    """
    The withName method returns the withName variable.

    RETURNS
    -------
    str
        The withName variable.
    """
    def withName(self) -> str:
        return self.__withName



from Dictionary.Word import Word

from MorphologicalAnalysis.FsmParse import FsmParse


class FsmParseList:
    __fsmParses: list

    """
    A constructor of FsmParseList class which takes a list fsmParses as an input. First it sorts
    the items of the list then loops through it, if the current item's transitions equal to the next item's
    transitions, it removes the latter item. At the end, it assigns this list to the fsmParses variable.

    PARAMETERS
    ----------
    fsmParses : list
        FsmParse list input.
    """

    def __init__(self, fsmParses: list):
        fsmParses.sort()
        i = 0
        while i < len(fsmParses) - 1:
            if fsmParses[i].transitionList() == fsmParses[i + 1].transitionList():
                fsmParses.pop(i + 1)
                i = i - 1
            i = i + 1
        self.__fsmParses = fsmParses

    """
    The size method returns the size of fsmParses list.

    RETURNS
    -------
    int
        The size of fsmParses list.
    """

    def size(self) -> int:
        return len(self.__fsmParses)

    """
    The getFsmParse method takes an integer index as an input and returns the item of fsmParses list at given index.

    PARAMETERS
    ----------
    index : int
        Integer input.
        
    RETURNS
    -------
    FsmParse
        The item of fsmParses list at given index.
    """

    def getFsmParse(self, index: int) -> FsmParse:
        return self.__fsmParses[index]

    """
    The rootWords method gets the first item's root of fsmParses {@link ArrayList} and uses it as currentRoot. Then loops through
    the fsmParses, if the current item's root does not equal to the currentRoot, it then assigns it as the currentRoot and
    accumulates root words in a {@link String} result.

    RETURNS
    -------
    str
        String result that has root words.
    """

    def rootWords(self) -> str:
        result = self.__fsmParses[0].getWord().getName()
        currentRoot = result
        for i in range(1, len(self.__fsmParses)):
            if self.__fsmParses[i].getWord().getName() != currentRoot:
                currentRoot = self.__fsmParses[i].getWord().getName()
                result = result + "$" + currentRoot
        return result

    """
    The reduceToParsesWithSameRootAndPos method takes a Word} currentWithPos as an input and loops i times till
    i equals to the size of the fsmParses list. If the given currentWithPos does not equal to the ith item's
    root and the MorphologicalTag of the first inflectional of fsmParses, it removes the ith item from the list.

    PARAMETERS
    ----------
    currentWithPos : Word
        Word input.
    """

    def reduceToParsesWithSameRootAndPos(self, currentWithPos: Word):
        i = 0
        while i < len(self.__fsmParses):
            if self.__fsmParses[i].getWordWithPos() != currentWithPos:
                self.__fsmParses.pop(i)
            else:
                i = i + 1

    """
    The reduceToParsesWithSameRoot method takes a str currentWithPos as an input and loops i times till
    i equals to the size of the fsmParses list. If the given currentRoot does not equal to the root of ith item of
    fsmParses, it removes the ith item from the list.

    PARAMETERS
    ----------
    currentRoot : str
        String input.
    """

    def reduceToParsesWithSameRoot(self, currentRoot: str):
        i = 0
        while i < len(self.__fsmParses):
            if self.__fsmParses[i].getWord().getName() == currentRoot:
                self.__fsmParses.pop(i)
            else:
                i = i + 1

    """
    The defaultCaseForParseString method takes String rootForm, parseString and partOfSpeech as inputs. And checks 
    defined cases for parseString and returns the strings till the $ sign. For example, if the given parseString is
    "A3PL+P3PL+NOM$A3PL+P3SG+NOM$A3PL+PNON+ACC$A3SG+P3PL+NOM" it returns "A3PL+P3SG+NOM".

    PARAMETERS
    ----------
    rootForm : str    
        String input.
    parseString : str
        String input.
    partOfSpeech : str
        String input.
        
    RETURNS
    -------
    str
        String defaultCase.
    """

    def defaultCaseForParseString(self, rootForm: str, parseString: str, partOfSpeech: str) -> str:
        if parseString == "P3SG+NOM$PNON+ACC":
            if partOfSpeech == "PROP":
                return "PNON+ACC"
            else:
                return "P3SG+NOM"
        elif parseString == "A2SG+P2SG$A3SG+P3SG":
            return "A3SG+P3SG"
        elif parseString == "A3PL+P3PL+NOM$A3PL+P3SG+NOM$A3PL+PNON+ACC$A3SG+P3PL+NOM":
            return "A3PL+P3SG+NOM"
        elif parseString == "P2SG$P3SG":
            return "P3SG"
        elif parseString == "A3PL+PNON+NOM$A3SG+PNON+NOM^DB+VERB+ZERO+PRES+A3PL":
            return "A3PL+PNON+NOM"
        elif parseString == "P2SG+NOM$PNON+GEN":
            return "PNON+GEN"
        elif parseString == "AOR^DB+ADJ+ZERO$AOR+A3SG":
            return "AOR+A3SG"
        elif parseString == "P2SG$PNON":
            return "PNON"
        elif parseString == "ADV+SINCE$VERB+ZERO+PRES+COP+A3SG":
            if rootForm == "yıl" or rootForm == "süre" or rootForm == "zaman" or rootForm == "ay":
                return "ADV+SINCE"
            else:
                return "VERB+ZERO+PRES+COP+A3SG"
        elif parseString == "CONJ$VERB+POS+IMP+A2SG":
            return "CONJ"
        elif parseString == "NEG+IMP+A2SG$POS^DB+NOUN+INF2+A3SG+PNON+NOM":
            return "POS^DB+NOUN+INF2+A3SG+PNON+NOM"
        elif parseString == "NEG+OPT+A3SG$POS^DB+NOUN+INF2+A3SG+PNON+DAT":
            return "POS^DB+NOUN+INF2+A3SG+PNON+DAT"
        elif parseString == "NOUN+A3SG+P3SG+NOM$NOUN^DB+ADJ+ALMOST":
            return "NOUN+A3SG+P3SG+NOM"
        elif parseString == "ADJ$VERB+POS+IMP+A2SG":
            return "ADJ"
        elif parseString == "NOUN+A3SG+PNON+NOM$VERB+POS+IMP+A2SG":
            return "NOUN+A3SG+PNON+NOM"
        elif parseString == "INF2+A3SG+P3SG+NOM$INF2^DB+ADJ+ALMOST$":
            return "INF2+A3SG+P3SG+NOM"
        else:
            return None

    """
    The caseDisambiguator method first calls the parsesWithoutPrefixAndSuffix method and gets the words without prefixes 
    and suffixes. If the size of fsmParses list is 1, it directly returns the first item of that list and null if
    the size is 0.

    Then, it calls defaultCaseForParseString method with the root of first item of fsmParses, result of 
    parsesWithoutPrefixAndSuffix method, and the pos of the first item and assigns it result to the defaultCase str. If 
    defaultCase is not null, it then loops through the fsmParses and checks whether the current transitionList of 
    FsmParse contains the defaultCase, if so it returns current FsmParse, null otherwise.

    RETURNS
    -------
    FsmParse
        FsmParse if it contains defaultCase, None otherwise.
    """

    def caseDisambiguator(self) -> FsmParse:
        parseString = self.parsesWithoutPrefixAndSuffix()
        if len(self.__fsmParses) == 1:
            return self.__fsmParses[0]
        if len(self.__fsmParses) == 0:
            return None
        defaultCase = self.defaultCaseForParseString(self.__fsmParses[0].getWord().getName(), parseString,
                                                     self.__fsmParses[0].getFinalPos())
        if defaultCase is not None:
            for i in range(len(self.__fsmParses)):
                fsmParse = self.__fsmParses[i]
                if defaultCase in fsmParse.transitionList():
                    return fsmParse
        return None

    """
    The constructParseListForDifferentRootWithPos method initially creates a result list then loops through the
    fsmParses list. For the first iteration, it creates new list as initial, then adds the
    first item od fsmParses to initial and also add this initial list to the result list. For the following iterations, 
    it checks whether the current item's root with the MorphologicalTag of the first inflectional
    equal to the previous item's  root with the MorphologicalTag of the first inflectional. If so, it adds that item
    to the result list, if not it creates new list as initial and adds the first item od fsmParses
    to initial and also add this initial list to the result list.

    RETURNS
    -------
    result : list
        list type of FsmParseList.
    """

    def constructParseListForDifferentRootWithPos(self) -> list:
        result = []
        i = 0
        while i < len(self.__fsmParses):
            if i == 0:
                initial = []
                initial.append(self.__fsmParses[i])
                result.append(FsmParseList(initial))
            else:
                if self.__fsmParses[i].getWordWithPos() == self.__fsmParses[i - 1].getWordWithPos():
                    result[len(result) - 1].__fsmParses.append(self.__fsmParses[i])
                else:
                    initial = []
                    initial.append(self.__fsmParses[i])
                    result.append(FsmParseList(initial))
            i = i + 1
        return result

    """
    The parsesWithoutPrefixAndSuffix method first creates a str array named analyses with the size of fsmParses list's 
    size.

    If the size is just 1, it then returns the first item's transitionList, if it is greater than 1, loops through the 
    fsmParses and puts the transitionList of each item to the analyses array.

    If the removePrefix condition holds, it loops through the analyses array and takes each item's substring after the 
    first + sign and updates that item of analyses array with that substring.

    If the removeSuffix condition holds, it loops through the analyses array and takes each item's substring till the 
    last + sign and updates that item of analyses array with that substring.

    It then removes the duplicate items of analyses array and returns a result str that has the accumulated items of 
    analyses array.

    RETURNS
    -------
    str
        result str that has the accumulated items of analyses array.
    """
    def parsesWithoutPrefixAndSuffix(self) -> str:
        analyses = [""] * len(self.__fsmParses)
        removePrefix = True
        removeSuffix = True
        if len(self.__fsmParses) == 1:
            return self.__fsmParses[0].transitionList()[self.__fsmParses[0].transitionList().index("+") + 1]
        for i in range(len(self.__fsmParses)):
            analyses[i] = self.__fsmParses[i].transitionList()
        while removePrefix:
            removePrefix = True
            for i in range(len(self.__fsmParses) - 1):
                if "+" not in analyses[i] or "+" not in analyses[i + 1] or \
                        analyses[i][:analyses[i].index("+") + 1] != analyses[i + 1][:analyses[i + 1].index("+") + 1]:
                    removePrefix = False
                    break
            if removePrefix:
                for i in range(len(self.__fsmParses)):
                    analyses[i] = analyses[i][analyses[i].index("+") + 1:]
        while removeSuffix:
            removeSuffix = True
            for i in range(len(self.__fsmParses) - 1):
                if "+" not in analyses[i] or "+" not in analyses[i + 1] or \
                        analyses[i][analyses[i].rindex("+") + 1] != analyses[i + 1][analyses[i + 1].rindex("+") + 1]:
                    removeSuffix = False
                    break
            if removeSuffix:
                for i in range(len(self.__fsmParses)):
                    analyses[i] = analyses[i][:analyses[i].rindex("+")]
        for i in range(len(analyses)):
            for j in range(i + 1, len(analyses)):
                if analyses[i] > analyses[j]:
                    analyses[i], analyses[j] = analyses[j], analyses[i]
        result = analyses[0]
        for i in range(len(analyses)):
            result = result + "$" + analyses[i]
        return result

    """
    The overridden toString method loops through the fsmParses list and accumulates the items to a result str.

    RETURNS
    -------
    str
        Result str that has the items of fsmParses list.
    """
    def __str__(self) -> str:
        result = ""
        for i in range(len(self.__fsmParses)):
            result += self.__fsmParses[i] + "\n"
        return result



from Dictionary.Word import Word

from MorphologicalAnalysis.InflectionalGroup import InflectionalGroup
from MorphologicalAnalysis.MorphologicalTag import MorphologicalTag


class MorphologicalParse:

    inflectionalGroups: list
    root: Word

    """
    Constructor of MorphologicalParse class which takes a String parse as an input. First it creates
    a list as iGs for inflectional groups, and while given String contains derivational boundary (^DB+), it
    adds the substring to the iGs list and continue to use given String from 4th index. If it does not contain ^DB+,
    it directly adds the given String to the iGs list. Then, it creates a new list as
    inflectionalGroups and checks for some cases.
    
    If the first item of iGs list is ++Punc, it creates a new root as +, and by calling
    InflectionalGroup method with Punc it initializes the IG list by parsing given input
    String IG by + and calling the getMorphologicalTag method with these substrings. If getMorphologicalTag method 
    returns a tag, it adds this tag to the IG list and also to the inflectionalGroups list.

    If the first item of iGs list has +, it creates a new word of first item's substring from index 0 to +,
    and assigns it to root. Then, by calling InflectionalGroup method with substring from index 0 to +,
    it initializes the IG list by parsing given input String IG by + and calling the getMorphologicalTag
    method with these substrings. If getMorphologicalTag method returns a tag, it adds this tag to the IG list
    and also to the inflectionalGroups list.

    If the first item of iGs list does not contain +, it creates a new word with first item and assigns it as root.
    At the end, it loops through the items of iGs and by calling InflectionalGroup method with these items
    it initializes the IG list by parsing given input String IG by + and calling the getMorphologicalTag
    method with these substrings. If getMorphologicalTag method returns a tag, it adds this tag to the IG list
    and also to the inflectionalGroups list.

    PARAMETERS
    ----------
    parse : str
        String input.
    """
    def __init__(self, parse:None):
        if parse is not None:
            if isinstance(parse, str):
                iGs = []
                st = parse
                while "^DB+" in st:
                    iGs.append(st[:st.index("^DB+")])
                    st = st[st.index("^DB+") + 4:]
                iGs.append(st)
                self.inflectionalGroups = []
                if iGs[0] == "++Punc":
                    self.root = Word("+")
                    self.inflectionalGroups.append(InflectionalGroup("Punc"))
                else:
                    if iGs[0].index("+") != -1:
                        self.root = Word(iGs[0][:iGs[0].index("+")])
                        self.inflectionalGroups.append(InflectionalGroup(iGs[0][iGs[0].index("+") + 1:]))
                    else:
                        self.root = Word(iGs[0])
                    for i in range(1, len(iGs)):
                        self.inflectionalGroups.append(InflectionalGroup(iGs[i]))
            elif isinstance(parse, list):
                self.inflectionalGroups = []
                if parse[0].index("+") != -1:
                    self.root = Word(parse[0][:parse[0].index("+")])
                    self.inflectionalGroups.append(InflectionalGroup(parse[0][parse[0].index("+") + 1:]))
                for i in range(1, len(parse)):
                    self.inflectionalGroups.append(InflectionalGroup(parse[i]))

    """
    The no-arg getWord method returns root Word.

    RETURNS
    -------
    Word
        root Word.
    """
    def getWord(self) -> Word:
        return self.root

    """
    The getTransitionList method gets the first item of inflectionalGroups list as a str, then loops
    through the items of inflectionalGroups and concatenates them by using +.

    RETURNS
    -------
    str
        String that contains transition list.
    """
    def getTransitionList(self) -> str:
        result = self.inflectionalGroups[0].__str__()
        for i in range(1, len(self.inflectionalGroups)):
            result = result + "+" + self.inflectionalGroups[i].__str__()
        return result

    """
    The getInflectionalGroupString method takes an int index as an input and if index is 0, it directly returns the
    root and the first item of inflectionalGroups list. If the index is not 0, it then returns the corresponding
    item of inflectionalGroups list as a str.

    PARAMETERS
    ----------
    index : int
        Integer input.
        
    RETURNS
    -------
    str
        Corresponding item of inflectionalGroups at given index as a str.
    """
    def getInflectionalGroupString(self, index: int) -> str:
        if index == 0:
            return self.root.getName() + "+" + self.inflectionalGroups[0].__str__()
        else:
            return self.inflectionalGroups[index].__str__()

    """
    The getInflectionalGroup method takes an integer index as an input and it directly returns the {InflectionalGroup
    at given index.

    PARAMETERS
    ----------
    index : int
        Integer input.
        
    RETURNS
    -------
    InflectionalGroup
        InflectionalGroup at given index.
    """
    def getInflectionalGroup(self, index: int) -> InflectionalGroup:
        return self.inflectionalGroups[index]

    """
    The getLastInflectionalGroup method directly returns the last InflectionalGroup of inflectionalGroups list.

    RETURNS
    -------
    InflectionalGroup
        The last InflectionalGroup of inflectionalGroups list.
    """
    def getLastInflectionalGroup(self) -> InflectionalGroup:
        return self.inflectionalGroups[len(self.inflectionalGroups) - 1]

    """
    The getTag method takes an integer index as an input and and if the given index is 0, it directly return the root.
    then, it loops through the inflectionalGroups list it returns the MorphologicalTag of the corresponding inflectional 
    group.

    PARAMETERS
    ----------
    index : int
        Integer input.
        
    RETURNS
    -------
    str
        The MorphologicalTag of the corresponding inflectional group, or null of invalid index inputs.
    """
    def getTag(self, index: int) -> str:
        size = 1
        if index == 0:
            return self.root.getName()
        for group in self.inflectionalGroups:
            if index < size + group.size():
                return InflectionalGroup.getTagString(group.getTag(index - size))
            size += group.size()
        return None

    """
    The tagSize method loops through the inflectionalGroups list and accumulates the sizes of each inflectional group
    in the inflectionalGroups.

    RETURNS
    -------
    int
        Total size of the inflectionalGroups list.
    """
    def tagSize(self) -> int:
        size = 1
        for group in self.inflectionalGroups:
            size += group.size()
        return size

    """
    The size method returns the size of the inflectionalGroups list.

    RETURNS
    -------
    int
        The size of the inflectionalGroups list.
    """
    def size(self) -> int:
        return len(self.inflectionalGroups)

    """
    The firstInflectionalGroup method returns the first inflectional group of inflectionalGroups list.

    RETURNS
    -------
    InflectionalGroup
        The first inflectional group of inflectionalGroups list.
    """
    def firstInflectionalGroup(self) -> InflectionalGroup:
        return self.inflectionalGroups[0]

    """
    The lastInflectionalGroup method returns the last inflectional group of inflectionalGroups list.

    RETURNS
    -------
    InflectionalGroup
        The last inflectional group of inflectionalGroups list.
    """
    def lastInflectionalGroup(self) -> InflectionalGroup:
        return self.inflectionalGroups[len(self.inflectionalGroups) - 1]

    """
    The getWordWithPos method returns root with the MorphologicalTag of the first inflectional as a new word.

    RETURNS
    -------
    Word
        Root with the MorphologicalTag of the first inflectional as a new word.
    """
    def getWordWithPos(self) -> Word:
        return Word(self.root.getName() + "+" + InflectionalGroup.getTagString(self.firstInflectionalGroup().getTag(0)))

    """
    The getPos method returns the MorphologicalTag of the last inflectional group.

    RETURNS
    -------
    str
        The MorphologicalTag of the last inflectional group.
    """
    def getPos(self) -> str:
        return InflectionalGroup.getTagString(self.lastInflectionalGroup().getTag(0))

    """
    The getRootPos method returns the MorphologicalTag of the first inflectional group.

    RETURNS
    -------
    str
        The MorphologicalTag of the first inflectional group.
    """
    def getRootPos(self) -> str:
        return InflectionalGroup.getTagString(self.firstInflectionalGroup().getTag(0))

    """
    The lastIGContainsCase method returns the MorphologicalTag of last inflectional group if it is one of the 
    NOMINATIVE, ACCUSATIVE, DATIVE, LOCATIVE or ABLATIVE cases, null otherwise.

    RETURNS
    -------
    str
        The MorphologicalTag of last inflectional group if it is one of the NOMINATIVE, ACCUSATIVE, DATIVE, LOCATIVE or 
        ABLATIVE cases, null otherwise.
    """
    def lastIGContainsCase(self) -> str:
        caseTag = self.lastInflectionalGroup().containsCase()
        if caseTag is not None:
            return InflectionalGroup.getTagString(caseTag)
        else:
            return "NULL"

    """
    The lastIGContainsTag method takes a MorphologicalTag as an input and returns true if the last inflectional group's
    MorphologicalTag matches with one of the tags in the IG list, false otherwise.

    PARAMETERS
    ----------
    tag : MorphologicalTag
        MorphologicalTag type input.
        
    RETURNS
    -------
    bool
        True if the last inflectional group's MorphologicalTag matches with one of the tags in the IG list, False 
        otherwise.
    """
    def lastIGContainsTag(self, tag: MorphologicalTag) -> bool:
        return self.lastInflectionalGroup().containsTag(tag)

    """
    lastIGContainsPossessive method returns true if the last inflectional group contains one of the
    possessives: P1PL, P1SG, P2PL, P2SG, P3PL AND P3SG, false otherwise.

    RETURNS
    -------
    bool
        True if the last inflectional group contains one of the possessives: P1PL, P1SG, P2PL, P2SG, P3PL AND P3SG, 
        false otherwise.
    """
    def lastIGContainsPossessive(self) -> bool:
        return self.lastInflectionalGroup().containsPossessive()

    """
    The isCapitalWord method returns True if the character at first index of root is an uppercase letter, False 
    otherwise.

    RETURNS
    -------
    bool
        True if the character at first index of root is an uppercase letter, False otherwise.
    """
    def isCapitalWord(self) -> bool:
        return self.root.getName()[0:0].isupper()

    """
    The isNoun method returns true if the past of speech is NOUN, False otherwise.

    RETURNS
    -------
    bool
        True if the past of speech is NOUN, False otherwise.
    """
    def isNoun(self) -> bool:
        return self.getPos() == "NOUN"

    """
    The isVerb method returns true if the past of speech is VERB, False otherwise.

    RETURNS
    -------
    bool
        True if the past of speech is VERB, False otherwise.
    """
    def isVerb(self) -> bool:
        return self.getPos() == "VERB"

    """
    The isRootVerb method returns True if the past of speech of root is BERV, False otherwise.

    RETURNS
    -------
    bool
        True if the past of speech of root is VERB, False otherwise.
    """
    def isRootVerb(self) -> bool:
        return self.getRootPos() == "VERB"

    """
    The isAdjective method returns True if the past of speech is ADJ, False otherwise.

    RETURNS
    -------
    bool
        True if the past of speech is ADJ, False otherwise.
    """
    def isAdjective(self) -> bool:
        return self.getPos() == "ADJ"

    """
    The isProperNoun method returns True if the first inflectional group's MorphologicalTag is a PROPERNOUN, False 
    otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a PROPERNOUN, False otherwise.
    """
    def isProperNoun(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.PROPERNOUN)

    """
    The isPunctuation method returns True if the first inflectional group's MorphologicalTag is a PUNCTUATION, False 
    otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a PUNCTUATION, False otherwise.
    """
    def isPunctuation(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.PUNCTUATION)

    """
    The isCardinal method returns True if the first inflectional group's MorphologicalTag is a CARDINAL, False 
    otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a CARDINAL, False otherwise.
    """
    def isCardinal(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.CARDINAL)

    """
    The isOrdinal method returns True if the first inflectional group's MorphologicalTag is a ORDINAL, False otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a ORDINAL, False otherwise.
    """
    def isOrdinal(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.ORDINAL)

    """
    The isReal method returns True if the first inflectional group's MorphologicalTag is a REAL, False otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a REAL, False otherwise.
    """
    def isReal(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.REAL)

    """
    The isNumber method returns True if the first inflectional group's MorphologicalTag is REAL or CARDINAL, False 
    otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a REAL or CARDINAL, False otherwise.
    """
    def isNumber(self) -> bool:
        return self.isReal() or self.isCardinal()

    """
    The isTime method returns True if the first inflectional group's MorphologicalTag is a TIME, False otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a TIME, False otherwise.
    """
    def isTime(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.TIME)

    """
    The isDate method returns True if the first inflectional group's MorphologicalTag is a DATE, False otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a DATE, False otherwise.
    """
    def isDate(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.DATE)

    """
    The isHashTag method returns True if the first inflectional group's MorphologicalTag is a HASHTAG, False otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a HASHTAG, False otherwise.
    """
    def isHashTag(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.HASHTAG)

    """
    The isEmail method returns True if the first inflectional group's MorphologicalTag is a EMAIL, False otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a EMAIL, False otherwise.
    """
    def isEmail(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.EMAIL)

    """
    The isPercent method returns True if the first inflectional group's MorphologicalTag is a PERCENT, False otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a PERCENT, False otherwise.
    """
    def isPercent(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.PERCENT)

    """
     The isFraction method returns True if the first inflectional group's MorphologicalTag is a FRACTION, False 
     otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a FRACTION, False otherwise.
    """
    def isFraction(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.FRACTION)

    """
    The isRange method returns True if the first inflectional group's MorphologicalTag is a RANGE, False otherwise.

    RETURNS
    -------
    bool
        True if the first inflectional group's MorphologicalTag is a RANGE, False otherwise.
    """
    def isRange(self) -> bool:
        return self.getInflectionalGroup(0).containsTag(MorphologicalTag.RANGE)

    """
    The isPlural method returns true if InflectionalGroup's MorphologicalTags are from the agreement plural
    or possessive plural, i.e A1PL, A2PL, A3PL, P1PL, P2PL or P3PL, and False otherwise.

    RETURNS
    -------
    bool
        True if InflectionalGroup's MorphologicalTags are from the agreement plural or possessive plural.
    """
    def isPlural(self) -> bool:
        for inflectionalGroup in self.inflectionalGroups:
            if inflectionalGroup.containsPlural():
                return True
        return False

    """
    The isAuxiliary method returns true if the root equals to the et, ol, or yap, and false otherwise.

    RETURNS
    -------
    bool
        True if the root equals to the et, ol, or yap, and False otherwise.
    """
    def isAuxiliary(self) -> bool:
        return self.root.getName() == "et" or self.root.getName() == "ol" or self.root.getName() == "yap"

    """
    The containsTag method takes a MorphologicalTag as an input and loops through the inflectionalGroups list,
    returns True if the input matches with on of the tags in the IG, False otherwise.

    PARAMETERS
    ----------
    tag : MorphologicalTag
        checked tag
        
    RETURNS
    -------
    bool
        True if the input matches with on of the tags in the IG, False otherwise.
    """
    def containsTag(self, tag: MorphologicalTag) -> bool:
        for inflectionalGroup in self.inflectionalGroups:
            if inflectionalGroup.containsTag(tag):
                return True
        return False

    """
    The getTreePos method returns the tree pos tag of a morphological analysis.

    RETURNS
    -------
    bool
        Tree pos tag of the morphological analysis in string form.
    """
    def getTreePos(self) -> str:
        if self.isProperNoun():
            return "NP"
        elif self.isVerb():
            return "VP"
        elif self.isAdjective():
            return "ADJP"
        elif self.isNoun():
            return "NP"
        elif self.containsTag(MorphologicalTag.ADVERB):
            return "ADVP"
        elif self.isCardinal():
            return "QP"
        elif self.containsTag(MorphologicalTag.POSTPOSITION):
            return "PP"
        elif self.containsTag(MorphologicalTag.CONJUNCTION):
            return "CONJP"
        elif self.containsTag(MorphologicalTag.DETERMINER):
            return "DP"
        elif self.containsTag(MorphologicalTag.INTERJECTION):
            return "INTJ"
        elif self.containsTag(MorphologicalTag.QUESTIONPRONOUN):
            return "WP"
        elif self.containsTag(MorphologicalTag.PRONOUN):
            return "NP"
        else:
            return "-XXX-"

    """
    The overridden toString method gets the root and the first inflectional group as a result String then concatenates
    with ^DB+ and the following inflectional groups.

    RETURNS
    -------
    str
        result str.
    """
    def __str__(self):
        result = self.root.getName() + "+" + self.inflectionalGroups[0].__str__()
        for i in range(1, len(self.inflectionalGroups)):
            result = result + "^DB+" + self.inflectionalGroups[i].__str__()
        return result



from setuptools import setup

setup(
    name='NlpToolkit-MorphologicalAnalysis',
    version='1.0.11',
    packages=['MorphologicalAnalysis'],
    url='https://github.com/olcaytaner/TurkishMorphologicalAnalysis-Py',
    license='',
    author='olcaytaner',
    author_email='olcaytaner@isikun.edu.tr',
    description='Turkish Morphological Analysis',
    install_requires=['NlpToolkit-Dictionary', 'NlpToolkit-Corpus', 'NlpToolkit-DataStructure']
)
